/*
 * 	My_Analyzer.h
 *
 *	This source file is under General Public License version 3. Please feel free to
 *	distribute it, hack it, or do anything else you like to it. I would ask, however
 *	that is you make any cool improvements, you let me know via any of the websites
 *	on which I have published this or by email at: WA2FZW@ARRL.net.
 *	
 *	This header file is used by the W8TEE/K2ZIA Antenna Analyzer in conjunction with
 *	the upgraded software by WA2FZW, with much help from G3ZQC, PE1PWF, K2RH and VK3PE.
 *	The definitions in here include the definitions of things needed to convert from
 *	the AD9850 DDS to the AD9851 DDS and things that allow the 6 meter band and "Custom"
 *	(i.e. any frequency range the owner wants) to be added or not.
 *
 *	In Version 02.5, we add support for Glenn's (VK3PE) AD8307 detector circuit in
 *	place of the germanium diodes.
 */

#ifndef My_Analyzer_h		// Prevent double #include
#define My_Analyzer_h


/*
 *	There are three optional configurations of the analyzer now. The original design uses
 *	the AD9850 DDS (Type II) board. Subsequent modifications were made to allow use of the
 *	AD9851 DDS, and most recently, the addition of support for the VK3PE AD8307 log
 *	amplifiers.
 *
 *	Before compiling for your particular analyzer, un-comment the definition for the DDS
 *	board you are using, and if using the AD8307 detector circuit, un-comment that as well.
 */

	#define	AD9850_DDS				// Using the AD9850 DDS board
//	#define AD9851_DDS				// Using the AD9851 DDS board

//	#define AD8307_SWR				// Using AD8307 detectors


/*
 *	Include all the necessary libraries. For the non-standard ones, use the
 *	links provided to find the ones that work with this code, as there are
 *	some different versions with identical or similar names that might not
 *	work correctly.
 */

#ifdef	AD9850_DDS				// If you have the AD9850 DDS
	#include <AD9850SPI.h>		// https://github.com/F4GOJ/AD9850SPI
#endif

#ifdef	AD9851_DDS				// If you are using the AD9851
	#include <SPI.h>			// Use the standard IDE library instead
#endif

#include <Adafruit_GFX.h>		// https://github.com/adafruit/Adafruit-GFX-Library
#include <MCUFRIEND_kbv.h>		// https://github.com/prenticedavid/MCUFRIEND_kbv
#include <Rotary.h>				// https://github.com/brianlow/Rotary
#include <EEPROM.h>				// Standard with IDE
#include <SD.h>					// Standard with IDE
#include <Wire.h>				// Standard with IDE
#include <math.h>				// Standard with IDE


/*
 *	The new code has all of the standard HF ham bands (160 meters through 10
 *	meters) permanently built in. However, if your analyzer uses the AD9851 DDS
 *	module, it is possible to add the 6 meter band as well, by enabling the
 *	definition of "ADD_6_METERS". Note, however that this is conditionalized
 *	upon whether or not "AD9851_DDS" is defined as 6 meters won't really work
 *	well without the higher frequency DDS.
 *
 *	If "ADD_6_METERS" is defined, you can also set the values of "LOW_6M_EDGE"
 *	and "HIGH_6M_EDGE" to values other than 50MHz to 54MHz as more useful
 *	limits might be more appropriate for what you do on 6 meters.
 */

#ifdef AD9851_DDS					// You can only add 6 meters if you have the AD9851

#define ADD_6_METERS				// Un-comment to enable 6 meter operation

	#define	LOW_6M_EDGE		50000U	// 50 MHz * 1000
	#define	HIGH_6M_EDGE	54000U	// 54 MHz * 1000

#endif


/*
 *	Regardless of which DDS module you have in your analyzer, you can add a "Custom"
 *	band definition. And as was the case with the inclusion of the 6 meter band, you
 *	can define the upper and lower limits however you like. We have them set for a
 *	range of 1MHz to 30MHz as the default, but be warned, that you're not going to
 *	see a lot of bandwidth detail on a graph of the SWR values (or maybe no dips at
 *	all) with that wide a range. Also if using the single frequency SWR option, it
 *	may take a few million turns of the encoder to get to where you want!
 *
 *	There is a more practical use for this option, however. If you're working with
 *	some kind of multiband antenna such as a tribander, you might want to set the
 *	band limits to 14MHz and 29.7MHz.
 *
 *	VERY IMPORTANT:	Be aware that when you use this feature to perform a full scan,
 *					you will be transmitting a low power signal across all frequencies
 *					between the band limits.
 */

#define ADD_CUSTOM					// Un-comment to enable the "Custom" band

#ifdef	ADD_CUSTOM					// If "Custom" is enabled

	#define LOW_C_EDGE	 1000U		//  1MHz * 1000
	#define	HIGH_C_EDGE	30000U		// 30MHz * 1000

#endif

/*
 *	These definitions are added in Version 02.8. They are used in the "RepeatScan()" and
 *	"DoNewScan()" functions.
 */
#define	DEFAULT_COUNT		 50				// Initial default repeat count
#define	REPEAT_INCREMENT	 10				// How much to change the count for each encoder click
#define	MIN_REPEAT_COUNT	 10				// Minimum repetition count (don't set any lower than 10)
#define	MAX_REPEAT_COUNT	100				// Maximum repetition count

#define	REPEAT_SCAN			  1				// For "DoNewScan()" - In repeat mode
#define	SINGLE_SCAN			  0				// For "DoNewScan()" - In single scan mode

#define SCAN_PAUSE		   1000				// Length of pause between scans (in mS)

bool	stopRepeating;						// For "DoNewScan()" - Stop repeating
int		repeatCount = DEFAULT_COUNT;		// How many times to repeat the scan


/*
 *	A slight modification in Version 02.9. Instead of the cute trick of using 2 pins tied
 *	together on the Arduino to detect the presence of the "Fine Tune" button, the folowing
 *	definition will tell the software whether or not it is installed. We discovered that for
 *	some reason, the button was drawing 80mA in the original configuration. That's twice
 *	the rating for a pin on the Arduino. Without the pins tied together, it draws about 10mA.
 *
 *	Comment out the following definition if you do not have the "Fine Tune" button installed.
 */

#define	FT_INSTALLED

#endif

